@@ -1,13 +1,23 @@
+from __future__ import division
+from __future__ import print_function
+
 __all__ = ['get']
-import os, curses
+import curses
+import os
+import sys
 
 cache = None
+
 def get(cap, *args, **kwargs):
     default = kwargs.pop('default', '')
 
     if 'PWNLIB_NOTERM' in os.environ:
         return ''
 
+    # Hack for readthedocs.org
+    if 'READTHEDOCS' in os.environ:
+        return ''
+
     if kwargs != {}:
         raise TypeError("get(): No such argument %r" % kwargs.popitem()[0])
 
@@ -36,6 +46,11 @@ def init():
     global cache
 
     if 'PWNLIB_NOTERM' not in os.environ:
-        curses.setupterm()
+        # Fix for BPython
+        try:
+            curses.setupterm()
+        except curses.error as e:
+            import traceback
+            print('Warning:', ''.join(traceback.format_exception_only(e.__class__, e)), file=sys.stderr)
 
     cache = {}
