@@ -1,10 +1,13 @@
+from __future__ import absolute_import
+from __future__ import division
+
 import ctypes
 import ctypes.util
 import socket
 
-from packing import pack
-from packing import p16
-from packing import p32
+from pwnlib.util.packing import p16
+from pwnlib.util.packing import p32
+from pwnlib.util.packing import pack
 
 __all__ = ['getifaddrs', 'interfaces', 'interfaces4', 'interfaces6', 'sockaddr']
 
@@ -223,4 +226,4 @@ def sockaddr(host, port, network = 'ipv4'):
         sockaddr += p32(0xffffffff) # Save three bytes 'push -1' vs 'push 0'
         sockaddr += host
         length    = len(sockaddr) + 4 # Save five bytes 'push 0'
-    return (sockaddr, length, address_family)
+    return (sockaddr, length, getattr(address_family, "name", address_family))
